-- # import nat
-- # import order

-- Non-standard variant of Fin: Fin n contains n + 1 elements.
-- Therefore Fin n is suitable for defining take for vectors of length
-- n, for instance. Note that Fin n is never empty.

data Fin : ℕ -> Set where
  fz (n : ℕ) : Fin n
  fs (n : ℕ) : Fin n -> Fin (suc n)

toNat (n : ℕ) : Fin n -> ℕ
toNat fz     = 0
toNat (fs i) = 1 + toNat i

prop_toNat (n : ℕ) : (i : Fin n) -> toNat i ≤ n
prop_toNat fz     = leZ
prop_toNat (fs i) = leS (prop_toNat i)

fromNat (n : ℕ) : (i : ℕ) -> i ≤ n -> Fin n
fromNat _ leZ      = fz
fromNat _ (leS le) = fs (fromNat _ le)

prop_fromNat_toNat (n : ℕ)
  : (i : Fin n) -> fromNat (toNat i) (prop_toNat i) ≡ i
prop_fromNat_toNat fz     = refl
prop_fromNat_toNat (fs i) = cong fs (prop_fromNat_toNat i)

prop_toNat_fromNat (i, n : ℕ)
  : (le : i ≤ n) -> toNat (fromNat i le) ≡ i
prop_toNat_fromNat leZ      = refl
prop_toNat_fromNat (leS le) = cong suc (prop_toNat_fromNat le)

raiseBound (n₁, n₂ : ℕ) : (i : Fin n₁) -> n₁ ≤ n₂ -> Fin n₂
raiseBound i le = fromNat (toNat i) $ prop_toNat i `transitive` le

changeBound (n₁, n₂ : ℕ) : (i : Fin n₂) -> toNat i ≤ n₁ -> Fin n₁
changeBound i = fromNat (toNat i)
