-- # import prelude
-- # import eq
-- # import monoid

data Maybe (a : Set) : Set where
  Just    : a -> Maybe a
  Nothing : Maybe a

mapMaybe (a, b : Set)
  : (a -> b) -> Maybe a -> Maybe b
mapMaybe f (Just x) = Just (f x)
mapMaybe f Nothing  = Nothing

fromMaybe (a : Set) : a -> Maybe a -> a
fromMaybe x Nothing  = x
fromMaybe _ (Just y) = y

isJust (a : Set) : (m : Maybe a) -> Set
isJust (Just _) = Unit
isJust Nothing  = Absurd

fromJust (a : Set) : (m : Maybe a) -> isJust m -> a
fromJust (Just x) unit = x

------------------------------------------------------------------------
-- Monoids

leftmost (a : Set) : Monoid (Maybe a)
leftmost = struct
  zero = Nothing
  (+)  = left
    where
    left (a : Set) : Maybe a -> Maybe a -> Maybe a
    left (Just x) _ = Just x
    left Nothing  y = y

  assoc = a
    where
    a (x, y, z : Maybe _) : x + (y + z) ≡ (x + y) + z
    a {Nothing} = refl
    a {Just _}  = refl

  leftUnit  = refl
  rightUnit = rU
    where
    rU (x : Maybe _) : x + zero ≡ x
    rU {Just _}  = refl
    rU {Nothing} = refl

rightmost (a : Set) : Monoid (Maybe a)
rightmost = struct
  zero = Nothing
  (+)  = right
    where
    right (a : Set) : Maybe a -> Maybe a -> Maybe a
    right _ (Just y) = Just y
    right x Nothing  = x

  assoc     = a
    where
    a (z, y, x : Maybe _) : x + (y + z) ≡ (x + y) + z
    a {Nothing} = refl
    a {Just _}  = refl

  leftUnit = lU
    where
    lU (x : Maybe _) : zero + x ≡ x
    lU {Just _}  = refl
    lU {Nothing} = refl

  rightUnit = refl

rightmostPreservesJust (a : Set) (m₁, m₂ : Maybe a)
  : isJust m₂ -> isJust (rightmost.(+) m₁ m₂)
rightmostPreservesJust {_} {_} {Just _} unit = unit

maybeMonoid₁ (a : Set) : (m : Monoid a) -> Monoid (Maybe a)
maybeMonoid₁ {a} m = struct
  zero = Nothing
  (+)  = f
    where
    f : (x, y : Maybe a) -> Maybe a
    f (Just x) (Just y) = Just (m.(+) x y)
    f Nothing  (Just y) = Just y
    f (Just x) Nothing  = Just x
    f Nothing  Nothing  = Nothing

  -- This one should be handled by QuickCheck-plugin, if possible.
  assoc = a
    where
    a (x, y, z : Maybe a) : x + (y + z) ≡ (x + y) + z
    a {Nothing} {Nothing} {Nothing} = refl
    a {Nothing} {Just _}  {Nothing} = refl
    a {Just _}  {Nothing} {Nothing} = refl
    a {Just _}  {Just _}  {Nothing} = refl
    a {Nothing} {Nothing} {Just _}  = refl
    a {Nothing} {Just _}  {Just _}  = refl
    a {Just _}  {Nothing} {Just _}  = refl
    a {Just _}  {Just _}  {Just _}  = cong Just m.assoc

  leftUnit = lU
    where
    lU (x : Maybe a) : zero + x ≡ x
    lU {Just _}  = refl
    lU {Nothing} = refl

  rightUnit = rU
    where
    rU (x : Maybe a) : x + zero ≡ x
    rU {Just _}  = refl
    rU {Nothing} = refl

maybeMonoid₂ (a : Set) : (m : Monoid a) -> Monoid (Maybe a)
maybeMonoid₂ {a} m = struct
  zero = Just m.zero
  (+)  = f
    where
    f : (x, y : Maybe a) -> Maybe a
    f (Just x) (Just y) = Just (m.(+) x y)
    f Nothing  (Just y) = Nothing
    f (Just x) Nothing  = Nothing
    f Nothing  Nothing  = Nothing

  -- This one should be handled by QuickCheck-plugin, if possible.
  assoc = a
    where
    a (x, y, z : Maybe a) : x + (y + z) ≡ (x + y) + z
    a {Nothing} {Nothing} {Nothing} = refl
    a {Nothing} {Just _}  {Nothing} = refl
    a {Just _}  {Nothing} {Nothing} = refl
    a {Just _}  {Just _}  {Nothing} = refl
    a {Nothing} {Nothing} {Just _}  = refl
    a {Nothing} {Just _}  {Just _}  = refl
    a {Just _}  {Nothing} {Just _}  = refl
    a {Just _}  {Just _}  {Just _}  = cong Just m.assoc

  leftUnit = lU
    where
    lU (x : Maybe a) : zero + x ≡ x
    lU {Just _}  = cong Just m.leftUnit
    lU {Nothing} = refl

  rightUnit = rU
    where
    rU (x : Maybe a) : x + zero ≡ x
    rU {Just _}  = cong Just m.rightUnit
    rU {Nothing} = refl

maybeMonoid₂Hom (a : Set) : (m : Monoid a) -> m ⇰ maybeMonoid₂ m
maybeMonoid₂Hom m = struct
  f    = Just
  zero = refl
  plus = \_ _ -> refl
