-- # import nat
-- # import order
-- # import prelude
-- # import monoid

infixr 5 ::

data Vec' (a : Set) : ℕ -> Set where
  nil : Vec' a zero
  (::) (n : ℕ) : a -> Vec' a n -> Vec' a (suc n)

Vec : (n : ℕ) -> (a : Set) -> Set
Vec n a = Vec' a n

data NonEmptyVec' (a : Set) : ℕ -> Set where
  singleton : a -> NonEmptyVec' a 0
  cons (n : ℕ) : a -> NonEmptyVec' a n -> NonEmptyVec' a (n + 1)

NonEmptyVec : (n : ℕ) -> (a : Set) -> Set
NonEmptyVec n a = NonEmptyVec' a n

------------------------------------------------------------------------

map (a, b : Set) (n : ℕ)
  : (a -> b) -> Vec n a -> Vec n b
map f nil       = nil
map f (x :: xs) = f x :: map f xs

mapNE (a, b : Set) (n : ℕ)
  : (a -> b) -> NonEmptyVec n a -> NonEmptyVec n b
mapNE f (singleton x) = singleton (f x)
mapNE f (cons x xs)   = cons (f x) (mapNE f xs)

foldr (a, b : Set) (n : ℕ)
  : (a -> b -> b) -> b -> Vec n a -> b
foldr f e nil       = e
foldr f e (x :: xs) = f x (foldr f e xs)

foldMonoid (a : Set) (n : ℕ)
  : Monoid a -> Vec n a -> a
foldMonoid m = foldr m.(+) m.zero

foldr1 (a : Set) (n : ℕ)
  : (a -> a -> a) -> Vec (n + 1) a -> a
foldr1 f (x :: nil)     = x
foldr1 f (x :: y :: xs) = x `f` foldr1 f (y :: xs)

replicate (n : ℕ) (a : Set)
  : a -> Vec n a
replicate {zero}  x = nil
replicate {suc n} x = x :: replicate {n} x

replicate1 (n : ℕ) (a : Set)
  : a -> a -> Vec n a
replicate1 {zero}  x y = nil
replicate1 {suc n} x y = x :: replicate {n} y

replicateNE (n : ℕ) (a : Set)
  : a -> NonEmptyVec n a
replicateNE {zero}  x = singleton x
replicateNE {suc n} x = cons x (replicateNE {n} x)

replicateNE1 (n : ℕ) (a : Set)
  : a -> a -> NonEmptyVec n a
replicateNE1 {zero}  x y = singleton x
replicateNE1 {suc n} x y = cons x (replicateNE {n} y)

prop_map_replicate (ℓ : ℕ) (a, b : Set) (x : a)
  : (f : a -> b) -> map f (replicate {ℓ} x) ≡ replicate {ℓ} (f x)
prop_map_replicate {0}     _ = refl
prop_map_replicate {suc ℓ} f = cong ((::) _) $ prop_map_replicate {ℓ} f

infixr 5 ++, +++

(++) (a : Set) (m, n : ℕ)
  : Vec m a -> Vec n a -> Vec (m + n) a
nil       ++ ys = ys
(x :: xs) ++ ys = x :: xs ++ ys

(+++) (a : Set) (m, n : ℕ)
  : Vec m a -> NonEmptyVec n a -> NonEmptyVec (m + n) a
nil       +++ ys = ys
(x :: xs) +++ ys = cons x (xs +++ ys)

appendRightNil (a : Set) (n : ℕ) : (xs : Vec n a) -> xs ++ [] ≡ xs
appendRightNil []        = refl
appendRightNil (x :: xs) = cong ((::) x) (appendRightNil xs)

appendRightNil'
  : (a : Set) -> (n : ℕ) -> (xs : Vec n a) -> xs ++ [] == xs
appendRightNil' _ _ _ = toPropEq (appendRightNil _)

rewrite appendRightNil'

prop_append_assoc (ℓ₁, ℓ₂, ℓ₃ : ℕ) (a : Set)
  :  (xs : Vec ℓ₁ a) -> (ys : Vec ℓ₂ a) -> (zs : Vec ℓ₃ a)
  -> subst (Vec' a) addAssociative (xs ++ (ys ++ zs)) ≡ (xs ++ ys) ++ zs
prop_append_assoc []        _  _  = refl
prop_append_assoc (x :: xs) ys zs =
  helper _ _ `trans` cong ((::) x) (prop_append_assoc xs ys zs)
  where
  helper (ℓ, ℓ' : ℕ) (a : Set) (x : a) (xs : Vec ℓ a)
    :  (eq₁ : ℓ + 1 ≡ ℓ' + 1) -> (eq₂ : ℓ ≡ ℓ')
    -> subst (Vec' a) eq₁ (x :: xs) ≡
       x :: subst (Vec' a) eq₂ xs
  helper refl refl = refl

prop_replicate_cons (m, n : ℕ) (a : Set) (x : a) (ys : Vec n a)
  : replicate {m} x ++ x :: ys ≡ replicate {m + 1} x ++ ys
prop_replicate_cons {0}     = refl
prop_replicate_cons {suc m} = cong ((::) _) prop_replicate_cons

prop_replicate_append (len1, len2 : ℕ) (a : Set) (x : a)
  : replicate {len1 + len2} x ≡ replicate {len1} x ++ replicate {len2} x
prop_replicate_append {0}        = refl
prop_replicate_append {suc len1} =
  cong ((::) _) (prop_replicate_append {len1})

prop_replicate_append' (len1, len2 : ℕ) (a : Set) (x : a)
  : replicateNE {len1 + len2} x ≡ replicate {len1} x +++ replicateNE {len2} x
prop_replicate_append' {0}        = refl
prop_replicate_append' {suc len1} =
  cong (cons _) (prop_replicate_append' {len1})

head (a : Set) (n : ℕ) : Vec (suc n) a -> a
head (x :: _) = x

headNE (a : Set) (n : ℕ) : NonEmptyVec n a -> a
headNE (singleton x) = x
headNE (cons x _)    = x

toHead (len : ℕ) (a : Set) : (a -> a) -> Vec len a -> Vec len a
toHead f nil       = nil
toHead f (x :: xs) = f x :: xs

toHeadNE (a : Set) (n : ℕ)
  : (a -> a) -> NonEmptyVec n a -> NonEmptyVec n a
toHeadNE f (singleton x) = singleton (f x)
toHeadNE f (cons x xs)   = cons (f x) xs

cast_toHead (a : Set) (n : ℕ) (xs : Vec n a) (f, g : a -> a)
  :  f ≡ g -> toHead f xs ≡ toHead g xs
cast_toHead refl = refl

prop_toHead_append_homomorphic
  (a : Set) (len, len' : ℕ)
  (xs : Vec (suc len) a) (ys : Vec len' a)
  (f : a -> a)
  : toHead f (xs ++ ys) ≡ toHead f xs ++ ys
prop_toHead_append_homomorphic {_} {_} {_} {_ :: _} {_} = refl

prop_toHead_functor_comp (a : Set) (n : ℕ)
  :  (f, g : a -> a) -> (xs : Vec n a)
  -> toHead f (toHead g xs) ≡ toHead (\x -> f (g x)) xs
prop_toHead_functor_comp _ _ nil      = refl
prop_toHead_functor_comp _ _ (_ :: _) = refl

prop_toHead_functor_id (a : Set) (n : ℕ)
  :  (xs : Vec n a)
  -> toHead (\x -> x) xs ≡ xs
prop_toHead_functor_id nil      = refl
prop_toHead_functor_id (_ :: _) = refl

prop_toHead_plus_assoc (len, x, y : ℕ) (xs : Vec len ℕ)
  : toHead ((+) x) (toHead ((+) y) xs) ≡ toHead ((+) (x + y)) xs
prop_toHead_plus_assoc =
  prop_toHead_functor_comp ((+) _) ((+) _) _
  `trans`
  cast_toHead addAssociative'

prop_toHead_plus_0 (len : ℕ) (xs : Vec len ℕ)
  : toHead ((+) 0) xs ≡ xs
prop_toHead_plus_0 = prop_toHead_functor_id _

prop_toHead_subtract_0 (len : ℕ) (xs : Vec len ℕ)
  : toHead (\n -> n ∸ 0) xs ≡ xs
prop_toHead_subtract_0 = prop_toHead_functor_id _

cast_toHeadNE (a : Set) (n : ℕ) (xs : NonEmptyVec n a) (f, g : a -> a)
  :  f ≡ g -> toHeadNE f xs ≡ toHeadNE g xs
cast_toHeadNE refl = refl

prop_toHeadNE_functor_comp (a : Set) (n : ℕ)
  :  (f, g : a -> a) -> (xs : NonEmptyVec n a)
  -> toHeadNE f (toHeadNE g xs) ≡ toHeadNE (\x -> f (g x)) xs
prop_toHeadNE_functor_comp _ _ (singleton _) = refl
prop_toHeadNE_functor_comp _ _ (cons _ _)    = refl

prop_toHeadNE_functor_id (a : Set) (n : ℕ)
  :  (xs : NonEmptyVec n a)
  -> toHeadNE (\x -> x) xs ≡ xs
prop_toHeadNE_functor_id (singleton _) = refl
prop_toHeadNE_functor_id (cons _ _)    = refl

prop_toHeadNE_plus_assoc (len, x, y : ℕ) (xs : NonEmptyVec len ℕ)
  : toHeadNE ((+) x) (toHeadNE ((+) y) xs) ≡ toHeadNE ((+) (x + y)) xs
prop_toHeadNE_plus_assoc =
  prop_toHeadNE_functor_comp ((+) _) ((+) _) _
  `trans`
  cast_toHeadNE addAssociative'

prop_toHeadNE_plus_0 (len : ℕ) (xs : NonEmptyVec len ℕ)
  : toHeadNE ((+) 0) xs ≡ xs
prop_toHeadNE_plus_0 = prop_toHeadNE_functor_id _

prop_toHeadNE_subtract_0 (len : ℕ) (xs : NonEmptyVec len ℕ)
  : toHeadNE (\n -> n ∸ 0) xs ≡ xs
prop_toHeadNE_subtract_0 = prop_toHeadNE_functor_id _

prop_toHeadNE_functor'
  (a : Set) (n : ℕ) (f : a -> a) (g : a -> a)
  (eq : (\x -> f (g x)) ≡ \x -> x)
  (xs : NonEmptyVec n a)
  : toHeadNE f (toHeadNE g xs) ≡ xs
prop_toHeadNE_functor' {_} {_} {_} {_} {eq} {singleton x} =
  cong (\h -> singleton (h x)) eq
prop_toHeadNE_functor' {_} {_} {_} {_} {eq} {cons x xs} =
  cong (\h -> cons (h x) xs) eq

prop_headNE_toHeadNE
  (a : Set) (len : ℕ) (xs : NonEmptyVec len a)
  : (f : a -> a) -> f (headNE xs) ≡ headNE (toHeadNE f xs)
prop_headNE_toHeadNE {_} {_} {singleton _} _ = refl
prop_headNE_toHeadNE {_} {_} {cons _ _}    _ = refl

tail (a : Set) (n : ℕ) : Vec (suc n) a -> Vec n a
tail (_ :: xs) = xs

fromVec (a : Set) (n : ℕ) : Vec (n + 1) a -> NonEmptyVec n a
fromVec (x :: nil)      = singleton x
fromVec (x :: x' :: xs) = cons x (fromVec (x' :: xs))

toVec (a : Set) (n : ℕ) : NonEmptyVec n a -> Vec (n + 1) a
toVec (singleton x) = x :: nil
toVec (cons x xs)   = x :: toVec xs

prop_fromVec_toVec (a : Set) (n : ℕ) (xs : NonEmptyVec n a)
  : fromVec (toVec xs) ≡ xs
prop_fromVec_toVec {_} {_} {singleton _}          = refl
prop_fromVec_toVec {_} {_} {cons _ (singleton _)} = refl
prop_fromVec_toVec {_} {_} {cons _ (cons _ _)}    =
  cong (cons _) (prop_fromVec_toVec {_} {_} {cons _ _})

prop_fromVec_toVec'
  :  (a : Set) -> (n : ℕ) -> (xs : NonEmptyVec n a)
  -> fromVec (toVec xs) == xs
prop_fromVec_toVec' _ _ _ = toPropEq prop_fromVec_toVec

prop_toVec_fromVec (a : Set) (n : ℕ) (xs : Vec (n + 1) a)
  : toVec (fromVec xs) ≡ xs
prop_toVec_fromVec {_} {_} {_ :: nil}    = refl
prop_toVec_fromVec {_} {_} {_ :: _ :: _} =
  cong ((::) _) prop_toVec_fromVec

prop_toVec_fromVec'
  :  (a : Set) -> (n : ℕ) -> (xs : Vec (n + 1) a)
  -> toVec (fromVec xs) == xs
prop_toVec_fromVec' _ _ _ = toPropEq prop_toVec_fromVec

prop_toVec_toHeadNE_toHead
  (a : Set) (f : a -> a) (n : ℕ) (xs : NonEmptyVec n a)
  : toVec (toHeadNE f xs) ≡ toHead f (toVec xs)
prop_toVec_toHeadNE_toHead {_} {_} {_} {singleton _} = refl
prop_toVec_toHeadNE_toHead {_} {_} {_} {cons _ _}    = refl

prop_headNE_fromVec (a : Set) (n : ℕ) (xs : Vec (n + 1) a)
  : headNE (fromVec xs) ≡ head xs
prop_headNE_fromVec {_} {_} {_ :: nil}    = refl
prop_headNE_fromVec {_} {_} {_ :: _ :: _} = refl

prop_headNE_fromVec_cons (a : Set) (n : ℕ) (x : a) (xs : Vec (n + 1) a)
  : headNE (fromVec (x :: xs)) ≡ x
prop_headNE_fromVec_cons {_} {_} {_} {_ :: nil}    = refl
prop_headNE_fromVec_cons {_} {_} {_} {_ :: _ :: _} = refl

prop_headNE_fromVec_cons'
  : (a : Set) -> (n : ℕ) -> (x : a) -> (xs : Vec (n + 1) a)
  -> headNE (fromVec (x :: xs)) == x
prop_headNE_fromVec_cons' _ _ _ _ = toPropEq prop_headNE_fromVec_cons

tailNE (a : Set) (n : ℕ) : NonEmptyVec (n + 1) a -> NonEmptyVec n a
tailNE (cons _ xs) = xs

tailNE' (a : Set) (n : ℕ) : NonEmptyVec n a -> Vec n a
tailNE' (singleton x) = nil
tailNE' (cons _ xs)   = toVec xs

prop_headNE_tailNE (a : Set) (n : ℕ) (xs : NonEmptyVec (n + 1) a)
  : cons (headNE xs) (tailNE xs) ≡ xs
prop_headNE_tailNE {_} {_} {cons _ _}  = refl

prop_headNE_tailNE' (a : Set) (n : ℕ) (xs : NonEmptyVec n a)
  : headNE xs :: tailNE' xs ≡ toVec xs
prop_headNE_tailNE' {_} {_} {singleton _} = refl
prop_headNE_tailNE' {_} {_} {cons _ _}    = refl

prop_tailNE'_toHeadNE
  (a : Set) (len : ℕ) (f : a -> a) (xs : NonEmptyVec len a)
  : tailNE' (toHeadNE f xs) ≡ tailNE' xs
prop_tailNE'_toHeadNE {_} {_} {_} {singleton _} = refl
prop_tailNE'_toHeadNE {_} {_} {_} {cons _ _}    = refl

prop_tailNE'_toHeadNE'
  :  (a : Set)
  -> (len : ℕ)
  -> (f : a -> a)
  -> (xs : NonEmptyVec len a)
  -> tailNE' (toHeadNE f xs) == tailNE' xs
prop_tailNE'_toHeadNE' _ _ _ _ = toPropEq prop_tailNE'_toHeadNE

prop_tailNE'_fromVec_cons (a : Set) (n : ℕ) (x : a) (xs : Vec n a)
  : tailNE' (fromVec (x :: xs)) ≡ xs
prop_tailNE'_fromVec_cons {_} {_}     {_} {nil}    = refl
prop_tailNE'_fromVec_cons {a} {suc n} {_} {_ :: _} =
  rewrite prop_toVec_fromVec' a n in refl

prop_tailNE'_fromVec_cons'
  : (a : Set) -> (n : ℕ) -> (x : a) -> (xs : Vec n a)
  -> tailNE' (fromVec (x :: xs)) == xs
prop_tailNE'_fromVec_cons' _ _ _ _ = toPropEq prop_tailNE'_fromVec_cons

take (ℓ : ℕ) (a : Set)
  : (i : ℕ) -> i ≤ ℓ -> Vec ℓ a -> Vec i a
take _ leZ      _         = []
take _ (leS le) (x :: xs) = x :: take _ le xs

prop_take_replicate (i, j : ℕ) (a : Set) (x : a)
  : (le : i ≤ j) -> take i le (replicate {j} x) ≡ replicate {i} x
prop_take_replicate leZ      = refl
prop_take_replicate (leS le) = cong ((::) _) $ prop_take_replicate le

prop_take_all (ℓ : ℕ) (a : Set)
  :  (xs : Vec ℓ a) -> take ℓ reflexive xs ≡ xs
prop_take_all []        = refl
prop_take_all (x :: xs) = cong ((::) x) $ prop_take_all xs

prop_take_append₁ (ℓ₁, ℓ₂ : ℕ) (a : Set)
  :  (i : ℕ) -> (le : i ≤ ℓ₁) -> (xs : Vec ℓ₁ a) -> (ys : Vec ℓ₂ a)
  -> take i (plusRightIncreasing le) (xs ++ ys) ≡ take i le xs
prop_take_append₁ 0 leZ xs ys = refl
prop_take_append₁ {suc ℓ₁} {ℓ₂} (suc i) (leS le) (x :: xs) ys =
  cong ((::) x) $ prop_take_append₁ i le xs ys

prop_take_append₂ (ℓ₁, ℓ₂ : ℕ) (a : Set)
  :  (i : ℕ) -> (le₁ : ℓ₁ ≤ i) -> (le₂ : i ≤ ℓ₁ + ℓ₂)
  -> (xs : Vec ℓ₁ a) -> (ys : Vec ℓ₂ a)
  -> take i le₂ (xs ++ ys) ≡
     subst (Vec' a) (prop_plus_minus le₁)
       (xs ++ take (minus le₁) (prop_minus_plus le₂ le₁) ys)
prop_take_append₂ i leZ le₂ [] ys = refl
prop_take_append₂ (suc i) (leS le₁) (leS le₂) (x :: xs) ys =
  cong ((::) x) (prop_take_append₂ i le₁ le₂ xs ys) `trans` helper _
  where
  helper (a : Set) (ℓ, ℓ' : ℕ) (z : a) (zs : Vec ℓ a)
    : (eq : ℓ ≡ ℓ')
    -> z :: subst (Vec' a) eq zs ≡ subst (Vec' a) (cong suc eq) (z :: zs)
  helper refl = refl

zipWith (a, b, c : Set) (len : ℕ)
  : (a -> b -> c) -> Vec len a -> Vec len b -> Vec len c
zipWith f nil       nil       = nil
zipWith f (x :: xs) (y :: ys) = f x y :: zipWith f xs ys

zipWithNE (a, b, c : Set) (len : ℕ)
  :  (a -> b -> c)
  -> NonEmptyVec len a -> NonEmptyVec len b -> NonEmptyVec len c
zipWithNE f (singleton x) (singleton y) = singleton (f x y)
zipWithNE f (cons x xs)   (cons y ys)   =
  cons (f x y) (zipWithNE f xs ys)

enumFrom (len : ℕ) : ℕ -> Vec len ℕ
enumFrom {zero}    x = nil
enumFrom {suc len} x = x :: enumFrom (suc x)

enumFromNE (len : ℕ) : ℕ -> NonEmptyVec len ℕ
enumFromNE {zero}    x = singleton x
enumFromNE {suc len} x = cons x (enumFromNE (suc x))

enumFromThen (len : ℕ) : ℕ -> ℕ -> Vec len ℕ
enumFromThen {zero}    _ _     = nil
enumFromThen {suc len} x delta = x :: enumFromThen (x + delta) delta

enumFromThenNE (len : ℕ) : ℕ -> ℕ -> NonEmptyVec len ℕ
enumFromThenNE {zero}    x _     = singleton x
enumFromThenNE {suc len} x delta = cons x (enumFromThenNE (x + delta) delta)

sum (len : ℕ) : Vec len ℕ -> ℕ
sum nil       = 0
sum (x :: xs) = x + sum xs

prop_sum_append (ℓ₁, ℓ₂ : ℕ)
  :  (xs : Vec ℓ₁ ℕ) -> (ys : Vec ℓ₂ ℕ)
  -> sum (xs ++ ys) ≡ sum xs + sum ys
prop_sum_append []        ys = refl
prop_sum_append (x :: xs) ys =
  cong ((+) x) (prop_sum_append xs ys) `trans` addAssociative

prop_sum_replicate
  :  (i, n : ℕ) -> sum (replicate {i} n) ≡ i * n
prop_sum_replicate 0       n = refl
prop_sum_replicate (suc i) n = cong ((+) n) $ prop_sum_replicate i n

cumulativeSum' (len : ℕ) : ℕ -> Vec len ℕ -> Vec len ℕ
cumulativeSum' _ nil       = nil
cumulativeSum' s (n :: ns) = sn :: cumulativeSum' sn ns
  where sn = s + n

cumulativeSum (len : ℕ) : Vec len ℕ -> Vec len ℕ
cumulativeSum ns = cumulativeSum' 0 ns

minimum (len : ℕ) : NonEmptyVec len ℕ -> ℕ
minimum (singleton n) = n
minimum (cons n xs)   = n `min` minimum xs

init (a : Set) (n : ℕ) : Vec (suc n) a -> Vec n a
init (x :: nil)      = nil
init (x :: x' :: xs) = x :: init (x' :: xs)

initNE (a : Set) (n : ℕ) : NonEmptyVec n a -> Vec n a
initNE (singleton x) = nil
initNE (cons x xs)   = x :: initNE xs

prop_take_init (ℓ : ℕ) (a : Set)
  :  (i : ℕ) -> (le : i ≤ ℓ) -> (xs : Vec (ℓ + 1) a)
  -> take i (plusLeftIncreasing {1} le) xs ≡ take i le (init xs)
prop_take_init 0       leZ      _                = refl
prop_take_init (suc i) (leS le) (x₁ :: x₂ :: xs) =
  cong ((::) x₁) $ prop_take_init i le (x₂ :: xs)

prop_init_replicate (a : Set) (x : a)
  : (ℓ : ℕ) -> init (replicate {ℓ + 1} x) ≡ replicate {ℓ} x
prop_init_replicate 0       = refl
prop_init_replicate (suc ℓ) = cong ((::) _) $ prop_init_replicate ℓ

last (a : Set) (n : ℕ) : Vec (suc n) a -> a
last (x :: nil)      = x
last (x :: x' :: xs) = last (x' :: xs)

lastNE (a : Set) (n : ℕ) : NonEmptyVec n a -> a
lastNE (singleton x) = x
lastNE (cons x xs)   = lastNE xs

mapLast (a, b : Set) (n : ℕ)
  : b -> (a -> b) -> Vec (suc n) a -> Vec (suc n) b
-- mapLast y f xs = map f (init xs) ++ y :: nil
mapLast y f (x :: nil)      = y :: nil
mapLast y f (x :: x' :: xs) =
  f x :: mapLast y f (x' :: xs)

mapLastNE (a, b : Set) (n : ℕ)
  : b -> (a -> b) -> NonEmptyVec n a -> NonEmptyVec n b
-- mapLastNE y f xs = map f (initNE xs) +++ singleton y
mapLastNE y f (singleton x) = singleton y
mapLastNE y f (cons x xs)   = cons (f x) (mapLastNE y f xs)
