------------------------------------------------------------------------
-- Some boring supporting lemmas
------------------------------------------------------------------------

module Correct.Lemmas where

open import Data.Nat
open import Data.Nat.Properties
open import Data.Vec
open import Data.Fin
open ℕ-semiringSolver
open import Relation.Binary.PropositionalEquality

abstract

  lem₁ : forall i j k -> i + j + k ≡ j + (i + k)
  lem₁ i j k =
    prove (i ∷ j ∷ k ∷ []) (I :+ J :+ K) (J :+ (I :+ K)) ≡-refl
    where I = var zero; J = var (suc zero); K = var (suc (suc zero))

  lem₂ : forall i j k -> i + (j + k) ≡ i + k + j
  lem₂ i j k =
    prove (i ∷ j ∷ k ∷ []) (I :+ (J :+ K)) (I :+ K :+ J) ≡-refl
    where I = var zero; J = var (suc zero); K = var (suc (suc zero))

  lem₃ : forall i j k -> i + j + k ≡ i + (k + j)
  lem₃ i j k =
    prove (i ∷ j ∷ k ∷ []) (I :+ J :+ K) (I :+ (K :+ J)) ≡-refl
    where I = var zero; J = var (suc zero); K = var (suc (suc zero))

  lem₄ : forall i j k -> i + j + k ≡ i + (j + k)
  lem₄ i j k =
    prove (i ∷ j ∷ k ∷ []) (I :+ J :+ K) (I :+ (J :+ K)) ≡-refl
    where I = var zero; J = var (suc zero); K = var (suc (suc zero))

  lem₅ : forall i j k l -> i + (j + k) + l ≡ i + j + k + l
  lem₅ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ (J :+ K) :+ L)
          (I :+ J :+ K :+ L)
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₆ : forall i j k l -> i + j + (k + l) ≡ i + (j + k) + l
  lem₆ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ J :+ (K :+ L))
          (I :+ (J :+ K) :+ L)
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₇ : forall i j k l -> i + (j + k + l) ≡ k + l + (i + j)
  lem₇ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ (J :+ K :+ L))
          (K :+ L :+ (I :+ J))
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₈ :  forall a b c d e f
       -> (a + b + (c + d)) + (e + f) ≡ (f + d) + (e + a + b + c)
  lem₈ a b c d e f =
    prove (a ∷ b ∷ c ∷ d ∷ e ∷ f ∷ [])
          ((A :+ B :+ (C :+ D)) :+ (E :+ F))
          ((F :+ D) :+ (E :+ A :+ B :+ C))
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))
          E = var (suc (suc (suc (suc zero))))
          F = var (suc (suc (suc (suc (suc zero)))))

  lem₉ :  forall a b c d e
       -> a + (b + c + d + e) ≡ e + (b + (c + d)) + a
  lem₉ a b c d e =
    prove (a ∷ b ∷ c ∷ d ∷ e ∷ [])
          (A :+ (B :+ C :+ D :+ E))
          (E :+ (B :+ (C :+ D)) :+ A)
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))
          E = var (suc (suc (suc (suc zero))))

  lem₁₀ :  forall a b c d e
       -> a + (b + (c + d + e)) ≡ b + (e + c + d) + a
  lem₁₀ a b c d e =
    prove (a ∷ b ∷ c ∷ d ∷ e ∷ [])
          (A :+ (B :+ (C :+ D :+ E)))
          (B :+ (E :+ C :+ D) :+ A)
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))
          E = var (suc (suc (suc (suc zero))))

  lem₁₁ :  forall a b c d e
       -> a + (b + c + d) + e ≡ d + e + (a + b + c)
  lem₁₁ a b c d e =
    prove (a ∷ b ∷ c ∷ d ∷ e ∷ [])
          (A :+ (B :+ C :+ D) :+ E)
          (D :+ E :+ (A :+ B :+ C))
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))
          E = var (suc (suc (suc (suc zero))))

  lem₁₂ :  forall a b c d e f
       -> a + b + c + (d + e + f) ≡ f + c + (a + b + d + e)
  lem₁₂ a b c d e f =
    prove (a ∷ b ∷ c ∷ d ∷ e ∷ f ∷ [])
          (A :+ B :+ C :+ (D :+ E :+ F))
          (F :+ C :+ (A :+ B :+ D :+ E))
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))
          E = var (suc (suc (suc (suc zero))))
          F = var (suc (suc (suc (suc (suc zero)))))

  lem₁₃ : forall i j k l -> i + (j + (k + l)) ≡ l + j + (i + k)
  lem₁₃ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ (J :+ (K :+ L)))
          (L :+ J :+ (I :+ K))
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₁₄ : forall i j k -> i + (j + k) ≡ j + (i + k)
  lem₁₄ i j k =
    prove (i ∷ j ∷ k ∷ []) (I :+ (J :+ K)) (J :+ (I :+ K)) ≡-refl
    where I = var zero; J = var (suc zero); K = var (suc (suc zero))

  lem₁₅ : forall i j k l -> i + j + (k + l) ≡ i + (k + j) + l
  lem₁₅ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ J :+ (K :+ L))
          (I :+ (K :+ J) :+ L)
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₁₆ : forall i j -> i + j ≡ j + 0 + i
  lem₁₆ i j = prove (i ∷ j ∷ []) (I :+ J) (J :+ con 0 :+ I) ≡-refl
    where I = var zero; J = var (suc zero)

  lem₁₇ :  forall a b c d e
        -> a + (b + c + (d + 0 + e)) ≡ d + b + c + (a + e)
  lem₁₇ a b c d e =
    prove (a ∷ b ∷ c ∷ d ∷ e ∷ [])
          (A :+ (B :+ C :+ (D :+ con 0 :+ E)))
          (D :+ B :+ C :+ (A :+ E))
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))
          E = var (suc (suc (suc (suc zero))))

  lem₁₈ : forall i j k l -> i + (j + k + l) ≡ i + j + k + l
  lem₁₈ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ (J :+ K :+ L))
          (I :+ J :+ K :+ L)
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₁₉ :  forall a b c d
        -> a + b + (c + d) ≡ a + c + (b + d)
  lem₁₉ a b c d =
    prove (a ∷ b ∷ c ∷ d ∷ [])
          (A :+ B :+ (C :+ D))
          (A :+ C :+ (B :+ D))
          ≡-refl
    where A = var zero; B = var (suc zero); C = var (suc (suc zero))
          D = var (suc (suc (suc zero)))

  lem₂₀ : forall i j -> i + j ≡ i + 0 + j
  lem₂₀ i j = prove (i ∷ j ∷ []) (I :+ J) (I :+ con 0 :+ J) ≡-refl
    where I = var zero; J = var (suc zero)

  lem₂₁ : forall i j k l -> i + (j + k + l) ≡ j + (i + (k + l))
  lem₂₁ i j k l =
    prove (i ∷ j ∷ k ∷ l ∷ [])
          (I :+ (J :+ K :+ L))
          (J :+ (I :+ (K :+ L)))
          ≡-refl
    where I = var zero; J = var (suc zero)
          K = var (suc (suc zero)); L = var (suc (suc (suc zero)))

  lem₂₂ : forall i j -> i + j ≡ j + i
  lem₂₂ i j = prove (i ∷ j ∷ []) (I :+ J) (J :+ I) ≡-refl
    where I = var zero; J = var (suc zero)
