------------------------------------------------------------------------
-- The simple operational semantics is functional
------------------------------------------------------------------------

module OperationalSemantics.Functional where

open import MainResults
open import Kind
open import TypeSystem
open import Substitution
open VarSubst
open import Value
open import OperationalSemantics.Binding
open Heap↦
open import OperationalSemantics.Simple
open import Data.Nat
open import Relation.Binary.HeterogeneousEquality

------------------------------------------------------------------------
-- Functionality

-- And a variant for applications.

data _=•⇓=_
  {Γ : _} {σ τ : _}
  {Σ₁ : Heap (Γ ▻ σ)} {v₀ : Γ ⊢⇓ σ → τ}
  {Δ¹ : _} {ρ¹ : Γ ▻ σ ⇒ Δ¹}
  {Σ₂¹ : Heap Δ¹} {v¹ : Δ¹ ⊢⇓ τ} {n¹ : ℕ}
  (d¹ : Σ₁ ∣ v₀ •⇓ ρ¹ ⋉ Σ₂¹ ∣ v¹ √ n¹)
  {Δ² : _} {ρ² : Γ ▻ σ ⇒ Δ²}
  {Σ₂² : Heap Δ²} {v² : Δ² ⊢⇓ τ} {n² : ℕ}
  (d² : Σ₁ ∣ v₀ •⇓ ρ² ⋉ Σ₂² ∣ v² √ n²)
  : Set where
  results•-= : Δ¹ ≅ Δ² -> ρ¹ ≅ ρ² -> Σ₂¹ ≅ Σ₂² -> v¹ ≅ v² -> d¹ =•⇓= d²

-- The simple semantics is functional; if you start with a given term
-- in a given heap you will always get the same result.

mutual

  functional : Functionality

  -- Values.

  functional eval-λ       eval-λ       =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional eval-λ²      eval-λ²      =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional eval-z       eval-z       =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional eval-s       eval-s       =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional eval-natrec  eval-natrec  =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional eval-,       eval-,       =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional eval-uncurry eval-uncurry =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl

  -- Variables.

  functional (eval-vz d¹) (eval-vz d²)
    with functional d¹ d²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional (eval-vs d¹) (eval-vs d²)
    with functional d¹ d²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl

  -- Fixpoints.

  functional (eval-fix d¹) (eval-fix d²)
    with functional d¹ d²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl

  -- Applications.

  functional (eval-· d₁¹ d₂¹) (eval-· d₁² d₂²)
    with functional d₁¹ d₁²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl
    with functional• d₂¹ d₂²
  ... | results•-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results-= ≅-refl ≅-refl ≅-refl ≅-refl

  functional•
    :  forall {Γ : Ctxt simple} {σ τ Σ₁} {v₀ : Γ ⊢⇓ σ → τ}
              {Δ¹} {ρ¹ : Γ ▻ σ ⇒ Δ¹} {Σ₂¹ v¹ n¹}
              {Δ²} {ρ² : Γ ▻ σ ⇒ Δ²} {Σ₂² v² n²}
    -> (d¹ : Σ₁ ∣ v₀ •⇓ ρ¹ ⋉ Σ₂¹ ∣ v¹ √ n¹)
    -> (d² : Σ₁ ∣ v₀ •⇓ ρ² ⋉ Σ₂² ∣ v² √ n²)
    -> d¹ =•⇓= d²

  functional• eval-·-s eval-·-s =
    results•-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional• (eval-·-λ d¹) (eval-·-λ d²)
    with functional d¹ d²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results•-= ≅-refl ≅-refl ≅-refl ≅-refl
  functional• eval-·-λ² eval-·-λ² =
    results•-= ≅-refl ≅-refl ≅-refl ≅-refl

  -- Natrec.

  functional• (eval-·-natrec-zero d₁¹ d₂¹)
              (eval-·-natrec-zero d₁² d₂²)
    with functional d₁¹ d₁²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl
    with functional d₂¹ d₂²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results•-= ≅-refl ≅-refl ≅-refl ≅-refl

  functional• (eval-·-natrec-suc d₁¹ d₂¹)
              (eval-·-natrec-suc d₁² d₂²)
    with functional d₁¹ d₁²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl
    with functional d₂¹ d₂²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results•-= ≅-refl ≅-refl ≅-refl ≅-refl

  functional• (eval-·-natrec-zero d₁¹ d₂¹)
              (eval-·-natrec-suc  d₁² d₂²)
    with functional d₁¹ d₁²
  ... | results-= ≅-refl ≅-refl ≅-refl ()

  functional• (eval-·-natrec-suc  d₁¹ d₂¹)
              (eval-·-natrec-zero d₁² d₂²)
    with functional d₁¹ d₁²
  ... | results-= ≅-refl ≅-refl ≅-refl ()

  -- Uncurry.

  functional• (eval-·-uncurry d₁¹ d₂¹)
              (eval-·-uncurry d₁² d₂²)
    with functional d₁¹ d₁²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl
    with functional d₂¹ d₂²
  ... | results-= ≅-refl ≅-refl ≅-refl ≅-refl =
    results•-= ≅-refl ≅-refl ≅-refl ≅-refl
